import mimetypes
import os

from typing import (  # noqa: F401
    Optional,
    Text,
)

from facebook_sdk.exceptions import FacebookSDKException


class FacebookFile(object):
    def __init__(self, path):
        # tpye: (Text) -> None
        super(FacebookFile, self).__init__()
        self.path = path
        if not os.path.exists(self.path):
            raise FacebookSDKException('File does not exist.')

    def read(self):  # type: () -> bytes
        """
        Override read to open, read and close fd after reading.
        """
        with open(self.path, mode='rb') as f:
            return f.read()

    @property
    def mime_type(self):  # type: () -> Optional[Text]
        return mimetypes.guess_type(self.path)[0]

    @property
    def name(self):  # type: () -> Text
        return os.path.basename(self.path)



import re

import six
from six.moves.urllib.parse import (
    parse_qs,
    urlencode,
    urlparse,
    urlunparse,
)
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Text,
)


def force_slash_prefix(value):  # type: (Text) -> Text
    return '/' + value if not (value and str(value).startswith('/')) else value


def base_graph_url_endpoint(url_to_trim):  # type: (Text) -> Text
    return re.sub(r'^https://.+\.facebook\.com(/v.+?)?/', '/', url_to_trim)


def remove_params_from_url(url, params_to_remove):  # type: (Text, Iterable[Text]) -> Text
    parsed = urlparse(url)
    qd = parse_qs(parsed.query, keep_blank_values=True)
    filtered = dict((k, v) for k, v in qd.items() if k not in params_to_remove)
    newurl = urlunparse([
        parsed.scheme,
        parsed.netloc,
        parsed.path,
        parsed.params,
        urlencode(filtered, doseq=True),  # query string
        parsed.fragment
    ])

    return newurl


def get_params_from_url(url):  # type: (Text) -> Dict[Text, List[Text]]
    parsed = urlparse(url)
    qd = parse_qs(parsed.query, keep_blank_values=True)
    return qd


def convert_params_to_utf8(params):  # type: (Mapping[Any, Any]) -> Mapping[Any, Any]
    return {
        k: v.encode("utf-8") if isinstance(v, six.text_type) else v
        for k, v in params.items()
    }


def smart_text(value, encoding='utf-8', **kwargs):  # type: (Text, Text, Any) -> Text
    if isinstance(value, six.text_type):
        return value
    elif isinstance(value, six.binary_type):
        return value.decode(encoding=encoding, **kwargs)
    else:
        return six.text_type(value)



BASE_GRAPH_URL = 'https://graph.facebook.com'
BASE_AUTHORIZATION_URL = 'https://www.facebook.com'

GRAPH_VERSION_V2_5 = 'v2.5'
GRAPH_VERSION_V2_6 = 'v2.6'
GRAPH_VERSION_V2_7 = 'v2.7'
GRAPH_VERSION_V2_8 = 'v2.8'
GRAPH_VERSION_V2_9 = 'v2.9'
GRAPH_VERSION_V2_10 = 'v2.10'
GRAPH_VERSION_V2_11 = 'v2.11'
GRAPH_VERSION_V2_12 = 'v2.12'

ISO_8601_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S%z'

DEFAULT_GRAPH_VERSION = GRAPH_VERSION_V2_12
DEFAULT_REQUEST_TIMEOUT = 60

METHOD_DELETE = 'DELETE'
METHOD_GET = 'GET'
METHOD_POST = 'POST'
METHOD_PUT = 'PUT'



import os

from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Text,
    Tuple,
    Type,
    Union,
    cast,
)

from facebook_sdk.authentication import (
    AccessToken,
    OAuth2Client,
)
from facebook_sdk.client import FacebookClient
from facebook_sdk.constants import (
    DEFAULT_GRAPH_VERSION,
    METHOD_DELETE,
    METHOD_GET,
    METHOD_POST,
)
from facebook_sdk.exceptions import FacebookSDKException
from facebook_sdk.facebook_file import FacebookFile
from facebook_sdk.request import (
    FacebookBatchRequest,
    FacebookRequest,
)


if TYPE_CHECKING:
    from facebook_sdk.response import FacebookResponse, FacebookBatchResponse  # noqa: F401

APP_ID_ENV_NAME = 'FACEBOOK_APP_ID'
APP_SECRET_ENV_NAME = 'FACEBOOK_APP_SECRET'


class FacebookApp(object):
    def __init__(self, app_id, app_secret):
        # type: (Text, Text) -> None
        super(FacebookApp, self).__init__()
        self.app_id = app_id
        self.secret = app_secret

    def access_token(self):
        # type: () -> AccessToken
        from facebook_sdk.authentication import AccessToken
        return AccessToken(
            access_token='{app_id}|{secret}'.format(
                app_id=self.app_id,
                secret=self.secret,
            ),
        )


class Facebook(object):
    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Facebook, self).__init__()

        self.config = {
            'app_id': os.getenv(APP_ID_ENV_NAME, kwargs.get('app_id')),
            'app_secret': os.getenv(APP_ID_ENV_NAME, kwargs.get('app_secret')),
            'default_graph_version': kwargs.get('default_graph_version', DEFAULT_GRAPH_VERSION),
            'default_access_token': kwargs.get('default_access_token'),
        }

        if not self.config['app_id']:
            raise FacebookSDKException(
                'Required "app_id" key not supplied in config and could not find '
                'fallback environment variable "{app_id_env_name}"'.format(
                    app_id_env_name=APP_ID_ENV_NAME,
                )
            )
        if not self.config['app_secret']:
            raise FacebookSDKException(
                'Required "app_secret" key not supplied in config '
                'and could not find fallback environment variable "{app_secret_env_name}"'.format(
                    app_secret_env_name=APP_SECRET_ENV_NAME,
                )
            )

        self.default_graph_version = self.config.get('default_graph_version')

        if self.config.get('default_access_token'):
            self.set_default_access_token(cast(Text, self.config.get('default_access_token')))

        self.app = FacebookApp(
            app_id=cast(Text, self.config['app_id']),
            app_secret=cast(Text, self.config['app_secret']),
        )
        self.client = FacebookClient(request_timeout=kwargs.get('default_request_timeout'))
        self.oauth_client = OAuth2Client(
            app=self.app,
            client=self.client,
            graph_version=self.default_graph_version,
        )

    def request(
        self,
        method,  # type: Text
        endpoint,  # type: Text
        access_token=None,  # type: Optional[Text]
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
        timeout=None,  # type: Optional[int]
    ):
        # type: (...) -> FacebookRequest
        access_token = access_token or getattr(self, 'default_access_token', None)
        graph_version = graph_version or self.default_graph_version

        return FacebookRequest(
            app=self.app,
            method=method,
            access_token=access_token,
            endpoint=endpoint,
            params=params,
            headers=headers,
            graph_version=graph_version,
            timeout=timeout,
        )

    def send_request(
        self,
        method,  # type: Text
        endpoint,  # type: Text
        access_token=None,  # type: Optional[Text]
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
        timeout=None,  # type: Optional[int]
    ):
        # type: (...) -> FacebookResponse
        request = self.request(
            method=method,
            access_token=access_token,
            endpoint=endpoint,
            params=params,
            headers=headers,
            graph_version=graph_version,
            timeout=timeout,
        )
        response = self.send_facebook_request(request=request)

        return response

    def send_facebook_request(self, request):
        # type: (FacebookRequest) -> FacebookResponse
        return self.client.send_request(request=request)

    def send_batch_request(
        self,
        requests,  # type: Union[Iterable[FacebookRequest], Mapping[Text, FacebookRequest]]
        access_token=None,  # type: Optional[Text]
        graph_version=None,  # type: Optional[Text]
        timeout=None,  # type: Optional[int]
    ):
        # type: (...) -> FacebookBatchResponse
        access_token = access_token or getattr(self, 'default_access_token', None)
        graph_version = graph_version or self.default_graph_version

        batch_request = FacebookBatchRequest(
            app=self.app,
            requests=requests,
            access_token=access_token,
            graph_version=graph_version,
            timeout=timeout,
        )

        response = self.client.send_batch_request(batch_request=batch_request)
        return response

    def set_default_access_token(self, access_token):
        # type: (Union[Text, AccessToken]) -> None
        if isinstance(access_token, str):
            self.default_access_token = AccessToken(access_token=access_token)
        elif isinstance(access_token, AccessToken):
            self.default_access_token = access_token
        else:
            raise ValueError('The default access token must be of type "str" or AccessToken')

    def file_to_upload(self, path):
        # type: (Text) -> FacebookFile
        return FacebookFile(path=path)

    def post(
        self,
        endpoint,  # type: Text
        access_token=None,  # type: Optional[Text]
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
    ):
        return self.send_request(
            method=METHOD_POST,
            access_token=access_token,
            endpoint=endpoint,
            params=params,
            headers=headers,
            graph_version=graph_version,
        )

    def get(
        self,
        endpoint,  # type: Text
        access_token=None,  # type: Optional[Text]
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
    ):
        return self.send_request(
            method=METHOD_GET,
            access_token=access_token,
            endpoint=endpoint,
            params=params,
            headers=headers,
            graph_version=graph_version,
        )

    def delete(
        self,
        endpoint,  # type: Text
        access_token=None,  # type: Optional[Text]
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
    ):
        return self.send_request(
            method=METHOD_DELETE,
            access_token=access_token,
            endpoint=endpoint,
            params=params,
            headers=headers,
            graph_version=graph_version,
        )



import datetime
import hashlib
import hmac

from six.moves.urllib.parse import urlencode
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Text,
    Union,
)

from facebook_sdk import (
    __VERSION__ as VERSION,
    constants,
)
from facebook_sdk.exceptions import FacebookSDKException
from facebook_sdk.request import FacebookRequest


if TYPE_CHECKING:
    from facebook_sdk.client import FacebookClient  # noqa: F401
    from facebook_sdk.facebook import FacebookApp  # noqa: F401
    from facebook_sdk.response import FacebookResponse  # noqa: F401


class AccessToken(object):
    def __init__(self, access_token, expires_at=None):
        # type: (Text, Optional[datetime.datetime]) -> None
        super(AccessToken, self).__init__()

        self.access_token = access_token

        if expires_at:
            self.expires_at = expires_at

    def app_secret_proof(self, secret):  # type: (Text) -> Text
        return hmac.new(bytes(secret.encode('utf-8')), self.access_token.encode('utf-8'), hashlib.sha256).hexdigest()

    def is_app_access_token(self):  # type: () -> bool
        return len(self.access_token.split('|')) == 2

    def is_long_lived(self):  # type: () -> bool
        if getattr(self, 'expires_at', None):
            return self.expires_at > datetime.datetime.utcnow() + datetime.timedelta(hours=2)

        return self.is_app_access_token()

    def is_expired(self):  # type: () -> bool
        if getattr(self, 'expires_at', None):
            return self.expires_at < datetime.datetime.utcnow()

        return False

    def __str__(self):  # type: () -> Text
        return self.access_token


class OAuth2Client(object):
    def __init__(self, app, client, graph_version=None):
        # type: (FacebookApp, FacebookClient, Optional[Text]) -> None
        super(OAuth2Client, self).__init__()
        self.app = app
        self.client = client
        self.graph_version = graph_version or constants.DEFAULT_GRAPH_VERSION  # type: Text

    def get_authorization_url(self, redirect_url, state, scope=None, params=None):
        # type: (Text, Text, Optional[Iterable[Text]], Optional[Dict[Text, Text]]) -> Text
        """ Generates an authorization URL to begin the process of authenticating a user.

        :param redirect_url The callback URL to redirect to.
        :param scope        A list of permissions to request.
        :param state        The CSPRNG-generated CSRF value.
        :param params       a dict of parameters to generate URL.
        """
        _scope = scope or []  # type: Iterable[Text]
        _params = params or {}  # type: Dict[Text, Text]

        _params.update({
            'client_id': self.app.app_id,
            'state': state,
            'response_type': 'code',
            'sdk': 'facebook-python-sdk-{version}'.format(version=VERSION),
            'redirect_uri': redirect_url,
            'scope': ','.join(_scope),
        })

        return '{base_url}/{graph_version}/dialog/oauth?{query}'.format(
            base_url=constants.BASE_AUTHORIZATION_URL,
            graph_version=self.graph_version,
            query=urlencode(_params)
        )

    def debug_token(self, access_token):  # type: (Union[Text, AccessToken]) -> Dict
        """
        https://developers.facebook.com/docs/graph-api/reference/v2.8/debug_token

        :param access_token:

        :raise FacebookSDKException

        :return: the token metadata
        """
        params = {
            'input_token': str(access_token)
        }

        self.last_request = FacebookRequest(
            app=self.app,
            access_token=self.app.access_token(),
            method='GET',
            endpoint='/debug_token',
            params=params,
            graph_version=self.graph_version,
        )

        response = self.client.send_request(self.last_request)

        return response.json_body

    def get_access_token_from_code(self, code, redirect_uri=''):
        # type: (Text, Text) -> AccessToken
        params = {
            'code': code,
            'redirect_uri': redirect_uri
        }

        return self._request_an_access_token(params=params)

    def get_long_lived_access_token(self, access_token):
        # type: (Union[Text, AccessToken]) -> AccessToken
        """ Exchanges a short-lived access token with a long-lived access token.

        :param access_token:

        :raise FacebookSDKException

        :return an AccessToken
        """
        params = {
            'grant_type': 'fb_exchange_token',
            'fb_exchange_token': str(access_token),
        }

        return self._request_an_access_token(params=params)

    def get_code_from_long_lived_access_token(self, access_token, redirect_uri):
        # type: (Union[Text, AccessToken], Text) -> Text
        params = {
            'redirect_uri': redirect_uri,
        }

        response = self._send_request_with_client_params(
            '/oauth/client_code',
            params=params,
            access_token=access_token,
        )

        data = response.json_body

        if not data.get('code'):
            raise FacebookSDKException('Code was not returned from Graph.', 401)

        return data['code']

    def _send_request_with_client_params(self, endpoint, params, access_token=None):
        # type: (Text, Dict, Optional[Union[Text, AccessToken]]) -> FacebookResponse
        """
        :return FacebookResponse
        """
        params.update(self._get_client_params())

        access_token = access_token or self.app.access_token()

        self.last_request = FacebookRequest(
            app=self.app,
            access_token=str(access_token),
            method='GET',
            endpoint=endpoint,
            params=params,
            graph_version=self.graph_version,
        )

        return self.client.send_request(self.last_request)

    def _get_client_params(self):  # type: () -> Dict
        return {
            'client_id': self.app.app_id,
            'client_secret': self.app.secret
        }

    def _request_an_access_token(self, params):  # type: (Dict) -> AccessToken
        response = self._send_request_with_client_params(
            endpoint='/oauth/access_token',
            params=params,
        )
        data = response.json_body

        if 'access_token' not in data:
            raise FacebookSDKException('Access token was not returned from Graph.', 401)

        # For exchanging a short lived token with a long lived token.
        # The expiration time in seconds will be returned as "expires".
        expires_at = None
        if 'expires' in data:
            expires_at = datetime.datetime.utcfromtimestamp(int(data['expires']))
        elif 'expires_in' in data:
            # For exchanging a code for a short lived access token.
            # The expiration time in seconds will be returned as "expires_in".
            # See: https://developers.facebook.com/docs/facebook-login/access-tokens#long-via-code
            expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=(data['expires_in']))

        return AccessToken(
            access_token=data['access_token'],
            expires_at=expires_at,
        )



from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Text,
    Tuple,
    Type,
    Union,
)


if TYPE_CHECKING:
    from facebook_sdk.response import FacebookResponse  # noqa: F401


SUB_CODE_AUTH_EXCEPTION_CODES = (458, 459, 460, 463, 464, 467)
SUB_CODE_RESUMABLE_UPLOAD_EXCEPTION_CODES = (1363030, 1363030, 1363037, 1363033, 1363021, 1363041)
AUTH_EXCEPTION_CODES = (100, 102, 190)
SERVER_EXCEPTION_CODES = (1, 2)
THROTTLE_EXCEPTION_CODES = (4, 17, 341)
CLIENT_EXCEPTION_CODES = (506,)


class FacebookSDKException(Exception):
    pass


class FacebookRequestException(FacebookSDKException):
    pass


class FacebookResponseException(FacebookSDKException):
    def __init__(self, response, code, message, **kwargs):
        # type: (FacebookResponse, int, Text,  Any) -> None
        super(FacebookResponseException, self).__init__(code, message)
        self.response = response
        self.code = code
        self.message = message

        self.error_subcode = kwargs.get('error_subcode')
        self.error_user_msg = kwargs.get('error_user_msg', '')
        self.error_user_title = kwargs.get('error_user_title', '')
        self.type = kwargs.get('error_type', '')

    @staticmethod
    def create(response):
        # type: (FacebookResponse) -> FacebookResponseException
        data = response.json_body
        error = (
            data
            if data.get('error', {}).get('code') is None and data.get('code') is not None
            else data.get('error', {})
        )

        code = error.get('code', -1)
        subcode = error.get('error_subcode', -1)
        message = error.get('message', 'Unknown error from Graph.')
        exception = FacebookOtherException  # type: Type[FacebookResponseException]
        if (
            subcode in SUB_CODE_AUTH_EXCEPTION_CODES or
            code in AUTH_EXCEPTION_CODES or
            error.get('type') == 'OAuthException'
        ):
            exception = FacebookAuthenticationException
        elif subcode in SUB_CODE_RESUMABLE_UPLOAD_EXCEPTION_CODES:
            exception = FacebookResumableUploadException
        elif code in SERVER_EXCEPTION_CODES:
            exception = FacebookServerException
        elif code in THROTTLE_EXCEPTION_CODES:
            exception = FacebookThrottleException
        elif code == 10 or 200 <= code <= 299:
            exception = FacebookAuthorizationException

        return exception(
            response=response,
            code=code,
            message=message,
            error_subcode=subcode,
            error_user_title=error.get('error_user_title', ''),
            error_user_msg=error.get('error_user_msg', ''),
            error_type=error.get('type', ''),
        )


class FacebookAuthenticationException(FacebookResponseException):
    pass


class FacebookResumableUploadException(FacebookResponseException):
    pass


class FacebookServerException(FacebookResponseException):
    pass


class FacebookThrottleException(FacebookResponseException):
    pass


class FacebookClientException(FacebookResponseException):
    pass


class FacebookAuthorizationException(FacebookResponseException):
    pass


class FacebookOtherException(FacebookResponseException):
    pass



from copy import copy
import json

from typing import (  # noqa: F401
    TYPE_CHECKING,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    SupportsAbs,
    Text,
    Tuple,
    TypeVar,
    Union,
)

from facebook_sdk.constants import METHOD_GET
from facebook_sdk.exceptions import (
    FacebookResponseException,
    FacebookSDKException,
)
from facebook_sdk.request import (  # noqa: F401
    FacebookBatchRequest,
    FacebookRequest,
)
from facebook_sdk.utils import (
    base_graph_url_endpoint,
    smart_text,
)


if TYPE_CHECKING:
    from mypy_extensions import TypedDict
    ResponseRecord = TypedDict('ResponseRecord', {
        'name': Text,
        'response': 'FacebookResponse'
    })

T = TypeVar('T', bound=FacebookRequest)


class BaseResponse(Generic[T]):

    def __init__(self, request):  # type (T) -> None
        self.request = request


class ResponsePaginationMixin(object):
    def next_page_request(self):  # type: () -> Optional[FacebookRequest]
        """ Return a FacebookRequest for the next page of the current Response

        :return: a FacebookRequest
        """
        return self._build_pagination_request('next')

    def previous_page_request(self):  # type: () -> Optional[FacebookRequest]
        """ Return a FacebookRequest for the previous page of the current Response

        :return: a FacebookRequest
        """
        return self._build_pagination_request('previous')

    def _build_pagination_request(self, direction):  # type: (Text) -> Optional[FacebookRequest]
        if self.request.method != METHOD_GET:  # type: ignore
            raise FacebookSDKException('You can only paginate on a GET request.', 720)

        pagination_url = self.json_body.get('paging', {}).get(direction)  # type: ignore

        request = None
        if pagination_url:
            request = copy(self.request)  # type: ignore
            request.endpoint = base_graph_url_endpoint(pagination_url)

        return request


class FacebookResponse(ResponsePaginationMixin, BaseResponse[FacebookRequest]):
    """ A Facebook Response

    """

    def __init__(
        self,
        request,  # type: FacebookRequest
        http_status_code,  # type: int
        body,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
    ):
        # type: (...) -> None
        super(FacebookResponse, self).__init__(request)
        self.body = body
        self.http_status_code = http_status_code
        self.headers = headers

        self._parse_body()

        if self.is_error:
            self._build_exception()

    @property
    def is_error(self):  # type: () -> bool
        """ Check if the response is an error."""
        return 'error' in self.json_body

    def _parse_body(self):  # type: () -> None
        """ Parse the raw response to json."""
        try:
            self.json_body = json.loads(smart_text(self.body))  # type: Dict
        except Exception:
            self.json_body = {}

    def raiseException(self):  # type: () -> None
        """ Raise the FacebookSDKException."""
        raise self.exception

    def _build_exception(self):  # type: () -> None
        self.exception = FacebookResponseException.create(response=self)   # type: FacebookResponseException


class FacebookBatchResponse(FacebookResponse, BaseResponse[FacebookBatchRequest]):
    """ A Facebook Batch Response"""

    def __init__(self, batch_request, batch_response):
        # type: (FacebookBatchRequest, FacebookResponse) -> None
        super(FacebookBatchResponse, self).__init__(
            request=batch_request,
            body=batch_response.body,
            http_status_code=batch_response.http_status_code,
            headers=batch_response.headers
        )

        self.responses = self.build_responses(self.json_body)  # type: List[ResponseRecord]

    def build_responses(self, json_body):  # type: (Dict) -> List[ResponseRecord]
        """ Parse the json_body to a set of FacebookResponse.

        :param json_body: parsed batch response
        """
        responses = []  # type: List[ResponseRecord]

        for index, response in enumerate(json_body):
            request_name = self.request.requests[index]['name']
            request = self.request.requests[index]['request']

            body = response.get('body')
            code = response.get('code')
            headers = response.get('headers')

            responses.insert(
                index,
                {
                    'name': request_name,
                    'response': FacebookResponse(
                        request=request,
                        body=body,
                        headers=headers,
                        http_status_code=code,
                    ),
                }
            )

        return responses

    def __iter__(self):  # type: () -> Iterable[ResponseRecord]
        return iter(self.responses)



import json
import uuid

from six.moves.urllib.parse import urlencode
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Text,
    Tuple,
    Union,
)

from facebook_sdk.constants import (
    DEFAULT_GRAPH_VERSION,
    METHOD_POST,
)
from facebook_sdk.exceptions import FacebookSDKException
from facebook_sdk.facebook_file import FacebookFile
from facebook_sdk.utils import (
    convert_params_to_utf8,
    force_slash_prefix,
    get_params_from_url,
    remove_params_from_url,
)


MAX_REQUEST_BY_BATCH = 50

if TYPE_CHECKING:
    from mypy_extensions import TypedDict
    from facebook_sdk.facebook import FacebookApp  # noqa: F401
    from facebook_sdk.authentication import AccessToken  # noqa: F401
    RequestRecord = TypedDict('RequestRecord', {
        'name': Text,
        'request': 'FacebookRequest',
        'attached_files': Text
    }, total=False)


class FacebookRequest(object):
    """ A Facebook Request.

    """

    def __init__(
        self,
        app=None,  # type: Optional[FacebookApp]
        access_token=None,  # type: Optional[Union[Text, AccessToken]]
        method=None,  # type: Optional[Text]
        endpoint='',  # type: Text
        params=None,  # type: Optional[Dict]
        headers=None,  # type: Optional[Dict]
        graph_version=None,  # type: Optional[Text]
        timeout=None,  # type: Optional[int]
    ):
        # type: (...) -> None

        super(FacebookRequest, self).__init__()

        # Default empty dicts for dict params.
        headers = headers or {}
        params = params or {}
        graph_version = graph_version or DEFAULT_GRAPH_VERSION

        self.app = app
        self.access_token = access_token  # type: ignore
        self.method = method
        self.endpoint = endpoint
        self.graph_version = graph_version  # type: Text
        self.headers = headers  # type: Dict
        self.params = params   # type: Dict
        self.timeout = timeout

    @property
    def endpoint(self):  # type: () -> Text
        return self._endpoint  # type: ignore

    @endpoint.setter
    def endpoint(self, value):  # type: (Text) -> None
        params = get_params_from_url(value)
        if 'access_token' in params:
            access_token = ''.join(params['access_token'])  # type: Text
            if not self.access_token:
                self.access_token = access_token
            elif self.access_token != access_token:
                raise FacebookSDKException(
                    'Access token mismatch. The access token provided in the FacebookRequest '
                    'and the one provided in the URL or POST params do not match.'
                )

        self._endpoint = remove_params_from_url(value, params_to_remove=['access_token', 'appsecret_proof'])

    @property
    def access_token(self):  # type: () -> Optional[Text]
        return self._access_token

    @access_token.setter
    def access_token(self, value):
        if value:
            self._access_token = str(value)
        else:
            self._access_token = None

    @property
    def params(self):  # type: () -> Dict
        """ The url params.

        :rtype: dict
        """
        params = self._params.copy() if self.method != METHOD_POST else {}
        if self.access_token:
            params.update(dict(access_token=self.access_token))

        return params

    @params.setter
    def params(self, value):  # type: (Dict) -> None
        if 'access_token' in value:
            self.access_token = value.get('access_token')

        value.pop('access_token', None)
        value.pop('appsecret_proof', None)

        self._extract_files_from_params(value)
        self._params = getattr(self, '_params', {})
        self._params.update(value)

    @property
    def post_params(self):  # type: () -> Optional[Dict]
        """ The post params.

        :rtype: object
        """
        if self.method == METHOD_POST:
            return self._params.copy()

        return None

    @property
    def url(self):  # type: () -> Text
        """ The relative url to the graph api.

        :rtype: str
        """
        return force_slash_prefix(self.graph_version) + force_slash_prefix(self.endpoint)

    @property
    def batch_url(self):  # type: () -> Text
        """ The relative url to the graph api.

        :rtype: str
        """
        params = self.params
        url = self.url

        if self.method != METHOD_POST and params:
            return '{url}?{encoded_params}'.format(
                url=url,
                encoded_params=urlencode(convert_params_to_utf8(self.params))
            )

        return url

    @property
    def url_encode_body(self):  # type: () -> Optional[Text]
        """ Convert the post params to a urlencoded str

        :rtype: str
        """
        params = self.post_params

        if not params:
            return None

        return urlencode(convert_params_to_utf8(params))

    def add_headers(self, headers):  # type: (Iterable[Dict]) -> None
        """ Append headers to the request.

        :param headers: a list of headers
        """
        for header in headers:
            self.headers.update(header)

    def _extract_files_from_params(self, value):  # type: (Dict) -> None
        self.files = {}  # type: Dict[Text, FacebookFile]
        for k, v in value.items():
            if isinstance(v, FacebookFile):
                self.files[k] = v

        for k in self.files.keys():
            value.pop(k)

    def contain_files(self):  # type: () -> bool
        return bool(self.files)

    def files_to_upload(self):  # type: () -> List[Tuple[Text, Tuple[Text, FacebookFile, Optional[Text]]]]
        return [(name, (_file.name, _file, _file.mime_type)) for name, _file in self.files.items()]


class FacebookBatchRequest(FacebookRequest):
    """ A Facebook Batch Request.

    """

    def __init__(
        self,
        app=None,  # type: Optional[FacebookApp]
        requests=None,  # type: Optional[Union[Iterable[FacebookRequest], Mapping[Text, FacebookRequest]]]
        access_token=None,  # type: Optional[Union[Text, AccessToken]]
        graph_version=None,  # type: Optional[Text]
        timeout=None,  # type:  Optional[int]
    ):
        # type: (...) -> None
        """
        :param requests: a list of FacebookRequest
        :param access_token: the access token for the batch request
        :param graph_version: the graph version for the batch request
        """
        graph_version = graph_version or DEFAULT_GRAPH_VERSION

        super(FacebookBatchRequest, self).__init__(
            app=app,
            access_token=access_token,
            graph_version=graph_version,
            method=METHOD_POST,
            endpoint='',
            timeout=timeout,
        )

        self.requests = []  # type: List[RequestRecord]

        if requests:
            self.add(request=requests)

    def add(self, request, name=None):
        # tpye: (Union[Iterable[FacebookRequest], Mapping[Text, FacebookRequest]], Optional[Text]) -> None
        """ Append a request or a set of requests to the batches.

        :param request: an instance, list or dict of FacebookRequest
        :param name: the name of the request. keep it as None if you provide a set of requests
        """

        if isinstance(request, list):
            for index, req in enumerate(request):
                self.add(req, index)
            return

        if isinstance(request, dict):
            for key, req in request.items():
                self.add(req, key)
            return

        if not isinstance(request, FacebookRequest):
            raise FacebookSDKException('Arguments must be of type dict, list or FacebookRequest.')

        self._add_access_token(request)

        request_to_add = {
            'name': str(name),
            'request': request,
        }  # type: RequestRecord

        attached_files = self.extract_file_attachments(request)
        if attached_files:
            request_to_add['attached_files'] = attached_files

        self.requests.append(request_to_add)

    def _add_access_token(self, request):  # type: (FacebookRequest) -> None
        """ Set the batch request access token to the request if it wasn't provided.

        :type request: FacebookRequest
        """

        if not request.access_token:
            access_token = self.access_token

            if not access_token:
                raise FacebookSDKException('Missing access token on FacebookRequest and FacebookBatchRequest')

            request.access_token = self.access_token

    def prepare_batch_request(self):  # type: () -> None
        params = {
            'batch': self.requests_to_json_str(),
            'include_headers': True,
        }
        self._params.update(params)

    def request_entity_to_batch_array(self, request, request_name, attached_files):
        # type: (FacebookRequest, Text, Optional[Text]) -> Dict
        """ Convert a FacebookRequest entity to a request batch representation.

        :param request: a FacebookRequest
        :param request_name: the request name
        :return: a dict with the representation of the request
        """

        batch = {
            'headers': request.headers,
            'method': request.method,
            'relative_url': request.batch_url,
        }

        encoded_body = request.url_encode_body
        if encoded_body:
            batch['body'] = encoded_body

        if request_name is not None:
            batch['name'] = request_name

        if request.access_token != self.access_token:
            batch['access_token'] = request.access_token

        if attached_files:
            batch['attached_files'] = attached_files

        return batch

    def requests_to_json_str(self):  # type: () -> Text
        """ Convert the requests to json."""
        json_requests = [
            self.request_entity_to_batch_array(
                request_name=request['name'],
                request=request['request'],
                attached_files=request.get('attached_files'),
            ) for request in self.requests
        ]

        return json.dumps(json_requests)

    def validate_batch_request_count(self):  # type: () -> None
        """ Validate the request count before sending them as a batch.

            :raise FacebookSDKException
        """
        requests_count = len(self.requests)

        if not requests_count:
            raise FacebookSDKException('Empty batch requests')
        if requests_count > MAX_REQUEST_BY_BATCH:
            raise FacebookSDKException('The limit of requests in batch is %d' % MAX_REQUEST_BY_BATCH)

    def extract_file_attachments(self, request):  # type: (FacebookRequest) -> Text
        """ Remove files from the request and return file names removed."""
        file_names = []
        for _file in request.files.values():
            file_name = uuid.uuid4().hex
            self.files[file_name] = _file
            file_names.append(file_name)

        request.files.clear()

        return ','.join(file_names)

    def __iter__(self):  # type: () -> Iterable[RequestRecord]
        return iter(self.requests)



import requests
from six.moves.urllib.parse import urlencode
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Text,
    Tuple,
    Union,
)

from facebook_sdk.constants import (
    BASE_GRAPH_URL,
    DEFAULT_REQUEST_TIMEOUT,
)
from facebook_sdk.response import (
    FacebookBatchResponse,
    FacebookResponse,
)
from facebook_sdk.utils import convert_params_to_utf8


if TYPE_CHECKING:
    from facebook_sdk.facebook_file import FacebookFile  # noqa: F401
    from facebook_sdk.request import FacebookRequest, FacebookBatchRequest  # noqa: F401


class FacebookClient(object):

    def __init__(self, request_timeout=None):
        # type: (Optional[int]) -> None
        self.timeout = request_timeout or DEFAULT_REQUEST_TIMEOUT  # tpye: int

    def _prepareRequest(self, request):
        # type: (FacebookRequest) -> Dict
        url = BASE_GRAPH_URL + request.url  # type: str

        data = None  # type: Optional[Union[Optional[Dict], Text]]
        if request.contain_files():
            if request.post_params:
                # Content-Type form-data will be provided by requests lib
                data = request.post_params
        else:
            request.add_headers([
                {'Content-Type': 'application/x-www-form-urlencoded'},
            ])
            if request.post_params:
                data = urlencode(convert_params_to_utf8(request.post_params))

        return dict(
            url=url,
            method=request.method,
            params=request.params,
            data=data,
            headers=request.headers,
            files=request.files_to_upload(),
            timeout=request.timeout or self.timeout,
        )

    def send_request(self, request):
        # type: (FacebookRequest) -> FacebookResponse
        request_params = self._prepareRequest(request)

        res = self.send(
            **request_params
        )

        response = FacebookResponse(
            request=request,
            body=res.content,
            http_status_code=res.status_code,
        )

        if response.is_error:
            response.raiseException()

        return response

    def send(
        self,
        data,  # type: Dict
        headers,  # type: Dict
        method,  # type: Text
        params,  # type: Dict
        url,  # type: Text
        files,  # type: List[Tuple[Text, Tuple[Text, FacebookFile, Text]]]
        timeout,  # type: int
    ):
        # type: (...) -> Any
        response = requests.request(
            method=method,
            url=url,
            headers=headers,
            params=params,
            data=data,
            files=files,
            timeout=timeout,
        )
        return response

    def send_batch_request(self, batch_request):
        # type: (FacebookBatchRequest) -> FacebookBatchResponse
        batch_request.validate_batch_request_count()
        batch_request.prepare_batch_request()
        batch_response = self.send_request(request=batch_request)

        response = FacebookBatchResponse(
            batch_request=batch_request,
            batch_response=batch_response,
        )

        return response



from setuptools import setup
from facebook_sdk import __VERSION__

readme = open('README.rst').read()

tests_require = [
    'coveralls',
    'pytest',
    'pytest-cov',
]

setup(
    name='facebook-py-sdk',
    version=__VERSION__,
    description='Facebook Python SDK',
    long_description=readme,
    license="MIT",
    author='Zeta Hernandez',
    author_email='zetahernandez@gmail.com',
    url='https://github.com/eventbrite/facebook-py-sdk',
    packages=['facebook_sdk'],
    package_data={'facebook_sdk': ['py.typed']},
    install_requires=[
        'requests >=0.8',
        'six',
        'typing;python_version<"3.5"'
    ],
    tests_require=tests_require,
    extras_require={
        'testing': tests_require,
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
